==> matrix-0.0.4-2.0.x.tar.gz: matrix-0.0.4-2.0.x/matrix.c <==
/*
 * The Atari Falcon Matrix Device Driver
 *
 * Authors: Tomas Berndtsson <tobe@lysator.liu.se>
 *          Fredrik Noring   <noring@lysator.liu.se>
 *          
 *
 * Written January 1997
 *
 * BUGS:
 *  - temporary implementation uses major number 60
 *    (later to be moved to major 14)
 *  - Incomplete dma buffer implementation
 *  - No support for other matrix connection other than mp2
 *
 * Copyright (C) 1997 Fredrik Noring & Tomas Berndtsson
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/malloc.h>	/* for kmalloc() and kfree() */
#include <linux/sched.h>	/* for struct wait_queue etc */
#include <linux/fcntl.h>	/* for O_NONBLOCK */
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/mm.h>

#include <asm/atariints.h>	/* IRQ_TYPE_PRIO */
#include <asm/segment.h>	/* For put_user and get_user */
#include <asm/atarihw.h>	/* For YM sound registers */
#include <asm/traps.h>
#include <asm/irq.h>		/* for struct pt_regs */

#include "config.h"
#include "matrix.h"
#include "hw.h"

#undef STATICALLY_ALLOCATED_BUFFERS	/* may help if you run out of memory */

#ifdef STATICALLY_ALLOCATED_BUFFERS
static char static_dma_buffer[BUFFER_SIZE];
#endif

#if DEBUG_LEVEL > 0
#define DEBUG(n, x, y...) \
({ \
	   if(n <= DEBUG_LEVEL) \
		   printk("%s: %s(): "x"\n", __FILE__, __FUNCTION__ , ## y); \
})
#else
#define DEBUG(n, x, y...)
#endif

#define FATAL(x, y...)	DEBUG(1, x , ## y)
#define ERROR(x, y...)	DEBUG(2, x , ## y)
#define INFO(x, y...)	DEBUG(3, x , ## y)
#define BLAH(x, y...)	DEBUG(4, x , ## y)

#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))

static struct {
	int in_use;
	int clock_source;
} matrix_device;

static struct {
	int size, first, state;
	char *buffer;
	unsigned long count, written;
} dma_write;

static void codec_on(void);
static void codec_off(void);
#ifdef 0
static void matrix_sndreset(void);
#endif
static void matrix_setup_dma(void);

/* incomplete so far */

static int matrix_read(struct inode *node, struct file *file,
		       char *buf, int count)
{
  return 0;
}

static int dma_copy_write(int s, const char *buf)
{
	long wrote;

	wrote = min(dma_write.count, dma_write.size/2);

	memcpy_fromfs(dma_write.buffer + s*dma_write.size/2,
		      buf+dma_write.written, wrote);

	dma_write.count -= wrote;
	dma_write.written += wrote;

	return 1;
}

static int matrix_write(struct inode *node, struct file *file,
			const char *buf, int count)
{
	char *dmaptr;

	INFO("called");

	/* Don't use the DMA twice */
	if (matrix_device.in_use)
		return -EFAULT;

	/* Don't do anything if nothing is to be done */
	if (!count) return 0;

	dma_write.count = count;
	dma_write.written = 0;      /* Nothing has been written yet */

	if(dma_write.first) {
	  dma_write.first = 0;
	  if (!dma_copy_write(0, buf))
	    return -EFAULT;
	  if (!dma_copy_write(1, buf))
	    return -EFAULT;
	  dma_write.state = 0;
	}

	matrix_device.in_use = 1;

	tt_dmasnd.ctrl = DMASND_CTRL_ON | DMASND_CTRL_REPEAT;

	while (dma_write.count > 0) {

	  dmaptr = (char*) DMASNDGetAdr();
	  while ((dmaptr-dma_write.buffer >= (dma_write.state+0)*dma_write.size/2) &&
		 (dmaptr-dma_write.buffer <  (dma_write.state+1)*dma_write.size/2))
	    {	/* Simulated interrupt */
	      current->state = TASK_INTERRUPTIBLE;
	      current->timeout = jiffies + 20;
	      schedule();

	      if (current->signal & ~current->blocked) {
		codec_off();
		matrix_device.in_use = 0;
		return dma_write.written;
	      }

	      dmaptr = (char*) DMASNDGetAdr();
	    }

	  if (!dma_copy_write(dma_write.state, buf)) {
	    codec_off();
	    matrix_device.in_use = 0;
	    return -EFAULT;
	  }

	  /* Switch section */
	  dma_write.state = (~dma_write.state)&1;
	}


        tt_dmasnd.ctrl &= DMASND_CTRL_ON;
	matrix_device.in_use = 0;
	return dma_write.written;
}

static int matrix_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	switch(cmd) {

	case MATRIX_GET_SNDSTATUS:
	  {
	  break;
	  }

	case MATRIX_SET_CLK_SRC:
	  INFO("MATRIX_SET_CLK_SRC");
	  matrix_device.clock_source = (int) arg;
	  break;

	case MATRIX_SET_LTATTEN:
	  {
	    int atten;
	    
	    INFO("MATRIX_SET_LTATTEN");
	    atten = (15 - max(min(arg,15), 0)) << 8;
	    tt_dmasnd.output_atten &= 0xf0ff;
	    tt_dmasnd.output_atten |= (u_short) atten;
	    break;
	  }
	case MATRIX_SET_RTATTEN:
	  {
	    int atten;
	    
	    INFO("MATRIX_SET_RTATTEN");
	    atten = (15 - max(min(arg,15), 0)) << 4;
	    tt_dmasnd.output_atten &= 0xff0f;
	    tt_dmasnd.output_atten |= (u_short) atten;
	    break;
	  }

	case MATRIX_INIT_DMA:
	  {
	    dma_write.first=1;
	    matrix_setup_dma();
	    break;
	  }

	case MATRIX_EXIT_DMA:
	  {

	    /* Wait for DMA */
	    tt_dmasnd.ctrl &= DMASND_CTRL_ON;
	    while (tt_dmasnd.ctrl & DMASND_CTRL_ON) {
	      current->state = TASK_INTERRUPTIBLE;
	      current->timeout = jiffies + 2;
	      schedule();
	    }

	    /* Shut down DMA */
	    codec_off();
/* 	    matrix_sndreset(); */
	    break;
	  }
	
	/* this ioctl should really calculate the bufferposition as well */
	/* however, with small buffers, this is enough */
	case MATRIX_GET_WRITTEN:
	  {
	    size_t *written = (size_t *)arg;
	    put_user(dma_write.written, written);
	    break;
	  }
	
	/* doesn't work yet, working on it! */
	case MATRIX_TEST_EXTERNAL:
	  {
	    int n, *ext=(int *)arg;
	    int src[] = { 42, 43, 44, 45, 46, 0 };
	    int dst[] = {  0,  0,  0,  0,  0, 0 };

	    tt_dmasnd.mode = DMASND_MODE_STEREO | DMASND_MODE_8BIT;
	    tt_dmasnd.dac_src = CODEC_SOURCE_MATRIX;

	    tt_dmasnd.ctrl = DMASND_CTRL_SELECT_RECORD;
	    DMASNDSetBase((long)dst);
	    DMASNDSetEnd((long)dst+(long)sizeof(dst));

	    tt_dmasnd.ctrl = DMASND_CTRL_SELECT_REPLAY;
	    DMASNDSetBase((long)src);
	    DMASNDSetEnd((long)src+(long)sizeof(src));

	    matrix.source = 0x622a;
	    matrix.destination = 0x0008;
	    matrix.external_frequency_divider = 0x01;
	    matrix.internal_frequency_divider = 0x00;

	    tt_dmasnd.ctrl = DMASND_CTRL_ON | DMASND_CTRL_RECORD_ON;
	    current->state = TASK_INTERRUPTIBLE;
	    current->timeout = jiffies + 100;
	    schedule();
	    codec_off();

	    for(n=0 ; src[n] && (src[n] == dst[n]) ; n++);
	    if(src[n]) {
	      put_user(0, ext);
	      INFO("MATRIX_TEST_EXTERNAL: External clock not found");
	    } else {
	      put_user(1, ext);
	      INFO("MATRIX_TEST_EXTERNAL: External clock found");
	    }

	  }

	}
	return 0;
}


static void codec_on(void)
{
	if (matrix_device.clock_source == MATRIX_CLK_SRC_EXT) {
		matrix.source = 0x63b4;
		matrix.destination = 0x2080;
		matrix.external_frequency_divider = 0x01;
		matrix.internal_frequency_divider = 0x01;

	} else if (matrix_device.clock_source == MATRIX_CLK_SRC_25MHZ) {
		matrix.source = 0x0194;
		matrix.destination = 0x2080;
		matrix.external_frequency_divider = 0x00;
		matrix.internal_frequency_divider = 0x01;

	} else if (matrix_device.clock_source == MATRIX_CLK_SRC_32MHZ) {

		/* NOTE! Settings are NOT correct! */
		matrix.source = 0x0194;
		matrix.destination = 0x2080;
		matrix.external_frequency_divider = 0x00;
		matrix.internal_frequency_divider = 0x01;

	}
	codec.tracks = 0x00;
}

static void codec_off(void)
{
	tt_dmasnd.ctrl = DMASND_CTRL_OFF;
	matrix.source = 0x0000;
	matrix.destination = 0x0000;
	matrix.external_frequency_divider = 0x01;
	matrix.internal_frequency_divider = 0x01;
	codec.tracks = 0x00;
}

#ifdef 0
/* this is how TOS does a sndstatus(1) */
static void matrix_sndreset(void)
{
// u_short tmp;

  __asm__(
	  "nop\n"
	  "nop\n"
	  );
//  tmp = tt_dmasnd.output_atten;
//  tmp &= 0xf00f;
//  tt_dmasnd.output_atten &= 0xf11f;
//  tt_dmasnd.output_atten &= 0xf22f;
//  tt_dmasnd.output_atten &= 0xf44f;
//  tt_dmasnd.output_atten &= 0xf88f;
    tt_dmasnd.adc_src |= 0x04;

  __asm__(
	  "nop\n"
	  "nop\n"
	  "nop\n"
	  );
//  tt_dmasnd.output_atten &= 0xf00f;
//  tt_dmasnd.output_atten &= tmp;
    *(u_short *)&tt_dmasnd.adc_src &= 0xf3ff;
  __asm__(
	  "nop\n"
	  "nop\n"
	  "nop\n"
	  "nop\n"
	  );
  tt_dmasnd.dac_src |= 0x08;
  tt_dmasnd.dac_src &= ~0x08;

  INFO("done");
}
#endif

static void matrix_setup_dma(void)
{
      INFO("stopping DMA");
      codec_off();

      /* sndstatus(SND_RESET); */
/*       matrix_sndreset(); */

      tt_dmasnd.int_ctrl = 0;

      /* setmode(STEREO16); */
      tt_dmasnd.mode = DMASND_MODE_STEREO | DMASND_MODE_16BIT;
                       /* | DMASND_MODE_50KHZ; */

      /* soundcmd(ADDERIN,MATIN); */
      tt_dmasnd.dac_src = CODEC_SOURCE_MATRIX;
      
      tt_dmasnd.ctrl = DMASND_CTRL_SELECT_REPLAY;
      DMASNDSetBase((long)dma_write.buffer);
      DMASNDSetEnd((long)dma_write.buffer+(long)BUFFER_SIZE);

      /* dsptristate(ENABLE,ENABLE); */
      /* devconnect(...); */
      codec_on();
 
  INFO("finished");
}

/* poll!! */
#ifdef 0
static int matrix_select(struct inode *inode, struct file *file, int sel_type,
			 select_table *wait)
{
	switch(sel_type) {
	case SEL_IN:		/* read */
		return 1;
	case SEL_OUT:		/* write */
		return 1;
	default:
		ERROR("unknown sel_type");
		return 1;
	}
}
#endif

static int matrix_open(struct inode *inode, struct file *file)
{
	INFO("called");

	// if (!matrix_device.in_use) {
	//	matrix_sndreset();
	//	matrix_device.clock_source = MATRIX_CLK_SRC_25MHZ;
	// }
	// matrix_device.in_use = 1;

	dma_write.first=1;
	MOD_INC_USE_COUNT;

	return 0;
}

static void matrix_release(struct inode *inode, struct file *file)
{
	INFO("called");
	
	// if (!matrix_device.in_use)
       	//	ERROR("device closed without being used?!?");
	// matrix_sndreset();
	// matrix_device.in_use = 0;

	MOD_DEC_USE_COUNT;
}

static struct file_operations matrix_fops = {
	NULL,    /* no special matrix_lseek */
	matrix_read,
	matrix_write,
	NULL,    /* no special matrix_readdir */
	NULL,    /* matrix_poll */
	matrix_ioctl,
	NULL,    /* no special matrix_mmap */
	matrix_open,
	matrix_release,
	NULL,    /* no special fsync */
	NULL,    /* no special fasync */
	NULL,    /* no special check_media_change */
	NULL     /* no special revalidate */
};

int init_module(void)
{
	int r;

	INFO("called");

	r = register_chrdev(MATRIX_MAJOR, "matrix", &matrix_fops);
	if(r < 0) {
		FATAL("unable to register matrix device. Error code %d", r);
		return -EIO;
	}

	matrix_device.in_use = 0;
	dma_write.size = BUFFER_SIZE;
  
	dma_write.buffer = 0;
#ifdef STATICALLY_ALLOCATED_BUFFERS
	dma_write.buffer = static_dma_buffer;
#else
	dma_write.buffer = kmalloc(BUFFER_SIZE, GFP_KERNEL);
	if (dma_write.buffer == 0) {
		FATAL("Out of memory while allocating dma buffer.");
		FATAL("Try recompiling with #define STATICALLY_ALLOCATED_BUFFERS");
		return -ENOMEM;
	}
#endif
  
	BLAH("dma_buffer=%p",dma_write.buffer);
    
	return 0;
}

void cleanup_module(void)
{
	INFO("called");

#ifndef STATICALLY_ALLOCATED_BUFFERS
	if (dma_write.buffer)
		kfree(dma_write.buffer);
#endif

	unregister_chrdev(MATRIX_MAJOR, "matrix");
}

==> matrix-0.0.5-2.0.x.tar.gz: matrix-0.0.5-2.0.x/matrix.c <==
/*
 * The Atari Falcon Matrix Device Driver
 *
 * Authors: Tomas Berndtsson <tobe@lysator.liu.se>
 *          Fredrik Noring   <noring@lysator.liu.se>
 *          
 *
 * Written January 1997
 *
 * BUGS:
 *  - temporary implementation uses major number 60
 *    (later to be moved to major 14)
 *  - Incomplete dma buffer implementation
 *  - No support for other matrix connection other than mp2
 *
 * Copyright (C) 1997 Fredrik Noring & Tomas Berndtsson
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/malloc.h>	/* for kmalloc() and kfree() */
#include <linux/sched.h>	/* for struct wait_queue etc */
#include <linux/fcntl.h>	/* for O_NONBLOCK */
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/mm.h>

#include <asm/segment.h>	/* For put_user and get_user */
#include <asm/atarihw.h>	/* For YM sound registers */
#include <asm/traps.h>
#include <asm/irq.h>		/* for struct pt_regs */

#include "config.h"
#include "matrix.h"
#include "hw.h"

#undef STATICALLY_ALLOCATED_BUFFERS	/* may help if you run out of memory */

#ifdef STATICALLY_ALLOCATED_BUFFERS
static char static_dma_buffer[BUFFER_SIZE];
#endif

#if DEBUG_LEVEL > 0
#define DEBUG(n, x, y...) \
({ \
	   if(n <= DEBUG_LEVEL) \
		   printk("%s: %s(): "x"\n", __FILE__, __FUNCTION__ , ## y); \
})
#else
#define DEBUG(n, x, y...)
#endif

#define FATAL(x, y...)	DEBUG(1, x , ## y)
#define ERROR(x, y...)	DEBUG(2, x , ## y)
#define INFO(x, y...)	DEBUG(3, x , ## y)
#define BLAH(x, y...)	DEBUG(4, x , ## y)

#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))

static struct {
	int in_use;
	int clock_source;
} matrix_device;

static struct {
	int size, first, state;
	char *buffer;
	unsigned long count, written;
} dma_write;

static void codec_on(void);
static void codec_off(void);
#ifdef 0
static void matrix_sndreset(void);
#endif
static void matrix_setup_dma(void);

/* incomplete so far */

static int matrix_read(struct inode *node, struct file *file,
		       char *buf, int count)
{
  return 0;
}

static int dma_copy_write(int s, const char *buf)
{
	long wrote;

	wrote = min(dma_write.count, dma_write.size/2);

	memcpy_fromfs(dma_write.buffer + s*dma_write.size/2,
		      buf+dma_write.written, wrote);

	/* fill last bit with zeros to prevent replaying some */
	if(wrote < dma_write.size/2)
		memset(dma_write.buffer + s*dma_write.size/2 + wrote, 0, 
		       dma_write.size/2 - wrote);

	dma_write.count -= wrote;
	dma_write.written += wrote;

	return 0;
}

static int matrix_write(struct inode *node, struct file *file,
			const char *buf, int count)
{
	char *dmaptr;

	INFO("called");

	/* Don't use the DMA twice */
	if (matrix_device.in_use)
		return -EFAULT;

	/* Don't do anything if nothing is to be done */
	if (!count) return 0;

	dma_write.count = count;
	dma_write.written = 0;      /* Nothing has been written yet */

	if(dma_write.first) {
	  dma_write.first = 0;
	  if (dma_copy_write(0, buf))
	    return -EFAULT;
	  if (dma_copy_write(1, buf))
	    return -EFAULT;
	  dma_write.state = 0;
	}

	matrix_device.in_use = 1;

	tt_dmasnd.ctrl = DMASND_CTRL_ON | DMASND_CTRL_REPEAT;

	while (dma_write.count > 0) {

	  dmaptr = (char*) DMASNDGetAdr();
	  while ((dmaptr-dma_write.buffer >= (dma_write.state+0)*dma_write.size/2) &&
		 (dmaptr-dma_write.buffer <  (dma_write.state+1)*dma_write.size/2))
	    {	/* Simulated interrupt */
	      current->state = TASK_INTERRUPTIBLE;
	      current->timeout = jiffies + 20;
	      schedule();

	      if (current->signal & ~current->blocked) {
		codec_off();
		matrix_device.in_use = 0;
		return dma_write.written;
	      }

	      dmaptr = (char*) DMASNDGetAdr();
	    }

	  if (dma_copy_write(dma_write.state, buf)) {
	    codec_off();
	    matrix_device.in_use = 0;
	    return -EFAULT;
	  }

	  /* Switch section */
	  dma_write.state = (~dma_write.state)&1;
	}


        tt_dmasnd.ctrl &= DMASND_CTRL_ON;
	matrix_device.in_use = 0;
	return dma_write.written;
}

static int matrix_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	switch(cmd) {

	case MATRIX_GET_SNDSTATUS:
	  {
	  break;
	  }

	case MATRIX_SET_CLK_SRC:
	  INFO("MATRIX_SET_CLK_SRC");
	  matrix_device.clock_source = (int) arg;
	  break;

	case MATRIX_SET_LTATTEN:
	  {
	    int atten;
	    
	    INFO("MATRIX_SET_LTATTEN");
	    atten = (15 - max(min(arg,15), 0)) << 8;
	    tt_dmasnd.output_atten &= 0xf0ff;
	    tt_dmasnd.output_atten |= (u_short) atten;
	    break;
	  }
	case MATRIX_SET_RTATTEN:
	  {
	    int atten;
	    
	    INFO("MATRIX_SET_RTATTEN");
	    atten = (15 - max(min(arg,15), 0)) << 4;
	    tt_dmasnd.output_atten &= 0xff0f;
	    tt_dmasnd.output_atten |= (u_short) atten;
	    break;
	  }

	case MATRIX_INIT_DMA:
	  {
	    dma_write.first=1;
	    matrix_setup_dma();
	    break;
	  }

	case MATRIX_EXIT_DMA:
	  {

	    /* Wait for DMA */
	    tt_dmasnd.ctrl &= DMASND_CTRL_ON;
	    while (tt_dmasnd.ctrl & DMASND_CTRL_ON) {
	      current->state = TASK_INTERRUPTIBLE;
	      current->timeout = jiffies + 2;
	      schedule();
	    }

	    /* Shut down DMA */
	    codec_off();
/* 	    matrix_sndreset(); */
	    break;
	  }
	
	/* this ioctl should really calculate the bufferposition as well */
	/* however, with small buffers, this is enough */
	case MATRIX_GET_WRITTEN:
	  {
	    size_t *written = (size_t *)arg;
	    put_user(dma_write.written, written);
	    break;
	  }
	
	/* doesn't work yet, working on it! */
	case MATRIX_TEST_EXTERNAL:
	  {
	    int n, *ext=(int *)arg;
	    int src[] = { 42, 43, 44, 45, 46, 0 };
	    int dst[] = {  0,  0,  0,  0,  0, 0 };

	    tt_dmasnd.mode = DMASND_MODE_STEREO | DMASND_MODE_8BIT;
	    tt_dmasnd.dac_src = CODEC_SOURCE_MATRIX;

	    tt_dmasnd.ctrl = DMASND_CTRL_SELECT_RECORD;
	    DMASNDSetBase((long)dst);
	    DMASNDSetEnd((long)dst+(long)sizeof(dst));

	    tt_dmasnd.ctrl = DMASND_CTRL_SELECT_REPLAY;
	    DMASNDSetBase((long)src);
	    DMASNDSetEnd((long)src+(long)sizeof(src));

	    matrix.source = 0x622a;
	    matrix.destination = 0x0008;
	    matrix.external_frequency_divider = 0x01;
	    matrix.internal_frequency_divider = 0x00;

	    tt_dmasnd.ctrl = DMASND_CTRL_ON | DMASND_CTRL_RECORD_ON;
	    current->state = TASK_INTERRUPTIBLE;
	    current->timeout = jiffies + 100;
	    schedule();
	    codec_off();

	    for(n=0 ; src[n] && (src[n] == dst[n]) ; n++);
	    if(src[n]) {
	      put_user(0, ext);
	      INFO("MATRIX_TEST_EXTERNAL: External clock not found");
	    } else {
	      put_user(1, ext);
	      INFO("MATRIX_TEST_EXTERNAL: External clock found");
	    }

	  }

	}
	return 0;
}


static void codec_on(void)
{
	if (matrix_device.clock_source == MATRIX_CLK_SRC_EXT) {
		matrix.source = 0x63b4;
		matrix.destination = 0x2080;
		matrix.external_frequency_divider = 0x01;
		matrix.internal_frequency_divider = 0x01;

	} else if (matrix_device.clock_source == MATRIX_CLK_SRC_25MHZ) {
		matrix.source = 0x0194;
		matrix.destination = 0x2080;
		matrix.external_frequency_divider = 0x00;
		matrix.internal_frequency_divider = 0x01;

	} else if (matrix_device.clock_source == MATRIX_CLK_SRC_32MHZ) {

		/* NOTE! Settings are NOT correct! */
		matrix.source = 0x0194;
		matrix.destination = 0x2080;
		matrix.external_frequency_divider = 0x00;
		matrix.internal_frequency_divider = 0x01;

	}
	codec.tracks = 0x00;
}

static void codec_off(void)
{
	tt_dmasnd.ctrl = DMASND_CTRL_OFF;
	matrix.source = 0x0000;
	matrix.destination = 0x0000;
	matrix.external_frequency_divider = 0x01;
	matrix.internal_frequency_divider = 0x01;
	codec.tracks = 0x00;
}

#ifdef 0
/* this is how TOS does a sndstatus(1) */
static void matrix_sndreset(void)
{
// u_short tmp;

  __asm__(
	  "nop\n"
	  "nop\n"
	  );
//  tmp = tt_dmasnd.output_atten;
//  tmp &= 0xf00f;
//  tt_dmasnd.output_atten &= 0xf11f;
//  tt_dmasnd.output_atten &= 0xf22f;
//  tt_dmasnd.output_atten &= 0xf44f;
//  tt_dmasnd.output_atten &= 0xf88f;
    tt_dmasnd.adc_src |= 0x04;

  __asm__(
	  "nop\n"
	  "nop\n"
	  "nop\n"
	  );
//  tt_dmasnd.output_atten &= 0xf00f;
//  tt_dmasnd.output_atten &= tmp;
    *(u_short *)&tt_dmasnd.adc_src &= 0xf3ff;
  __asm__(
	  "nop\n"
	  "nop\n"
	  "nop\n"
	  "nop\n"
	  );
  tt_dmasnd.dac_src |= 0x08;
  tt_dmasnd.dac_src &= ~0x08;

  INFO("done");
}
#endif

static void matrix_setup_dma(void)
{
      INFO("stopping DMA");
      codec_off();

      /* sndstatus(SND_RESET); */
/*       matrix_sndreset(); */

      tt_dmasnd.int_ctrl = 0;

      /* setmode(STEREO16); */
      tt_dmasnd.mode = DMASND_MODE_STEREO | DMASND_MODE_16BIT;
                       /* | DMASND_MODE_50KHZ; */

      /* soundcmd(ADDERIN,MATIN); */
      tt_dmasnd.dac_src = CODEC_SOURCE_MATRIX;
      
      tt_dmasnd.ctrl = DMASND_CTRL_SELECT_REPLAY;
      DMASNDSetBase((long)dma_write.buffer);
      DMASNDSetEnd((long)dma_write.buffer+(long)BUFFER_SIZE);

      /* dsptristate(ENABLE,ENABLE); */
      /* devconnect(...); */
      codec_on();
 
  INFO("finished");
}

/* poll!! */
#ifdef 0
static int matrix_select(struct inode *inode, struct file *file, int sel_type,
			 select_table *wait)
{
	switch(sel_type) {
	case SEL_IN:		/* read */
		return 1;
	case SEL_OUT:		/* write */
		return 1;
	default:
		ERROR("unknown sel_type");
		return 1;
	}
}
#endif

static int matrix_open(struct inode *inode, struct file *file)
{
	INFO("called");

	// if (!matrix_device.in_use) {
	//	matrix_sndreset();
	//	matrix_device.clock_source = MATRIX_CLK_SRC_25MHZ;
	// }
	// matrix_device.in_use = 1;

	dma_write.first=1;
	MOD_INC_USE_COUNT;

	return 0;
}

static void matrix_release(struct inode *inode, struct file *file)
{
	INFO("called");
	
	// if (!matrix_device.in_use)
       	//	ERROR("device closed without being used?!?");
	// matrix_sndreset();
	// matrix_device.in_use = 0;

	MOD_DEC_USE_COUNT;
}

static struct file_operations matrix_fops = {
	NULL,    /* no special matrix_lseek */
	matrix_read,
	matrix_write,
	NULL,    /* no special matrix_readdir */
	NULL,    /* matrix_poll */
	matrix_ioctl,
	NULL,    /* no special matrix_mmap */
	matrix_open,
	matrix_release,
	NULL,    /* no special fsync */
	NULL,    /* no special fasync */
	NULL,    /* no special check_media_change */
	NULL     /* no special revalidate */
};

int init_module(void)
{
	int r;

	INFO("called");

	r = register_chrdev(MATRIX_MAJOR, "matrix", &matrix_fops);
	if(r < 0) {
		FATAL("unable to register matrix device. Error code %d", r);
		return -EIO;
	}

	matrix_device.in_use = 0;
	dma_write.size = BUFFER_SIZE;
  
	dma_write.buffer = 0;
#ifdef STATICALLY_ALLOCATED_BUFFERS
	dma_write.buffer = static_dma_buffer;
#else
	dma_write.buffer = kmalloc(BUFFER_SIZE, GFP_KERNEL);
	if (dma_write.buffer == 0) {
		FATAL("Out of memory while allocating dma buffer.");
		FATAL("Try recompiling with #define STATICALLY_ALLOCATED_BUFFERS");
		return -ENOMEM;
	}
#endif
  

	BLAH("dma_buffer=%p",dma_write.buffer);

	printk("Matrix driver installed with buffer size %d\n", BUFFER_SIZE);
    
	return 0;
}

void cleanup_module(void)
{
	INFO("called");

#ifndef STATICALLY_ALLOCATED_BUFFERS
	if (dma_write.buffer)
		kfree(dma_write.buffer);
#endif

	unregister_chrdev(MATRIX_MAJOR, "matrix");
}
